#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "common.h"
#include "macros.h"
#include "failures.h"
#include "xmalloc.h"
#include "symbol.h"
#include "hash.h"
#include "as_tree.h"
#include "tree2text.h"

typedef void  (*travers_handler_t)(struct ast_node * );

static void travers_tree(struct ast_node *node);
static void traverse_id_node(struct ast_node *node);
static void traverse_op_node(struct ast_node *node);
static void traverse_const_node(struct ast_node *node);
static void traverse_access_node(struct ast_node *node);
static void traverse_assign_node(struct ast_node *node);
static void traverse_call_node(struct ast_node *node);
#ifdef _MEM_AREA
static void traverse_mem_area_node(struct ast_node *node);
#endif
static void traverse_if_node(struct ast_node *node);
static void traverse_for_node(struct ast_node *node);
static void traverse_while_node(struct ast_node *node);
static void traverse_return_node(struct ast_node *node);
static void traverse_break_node(struct ast_node *node);
static void traverse_continue_node(struct ast_node *node);
static void traverse_function_node(struct ast_node *node);
static void traverse_scope_node(struct ast_node *node);
static void traverse_scope_end_node(struct ast_node *node);
static void traverse_not_node(struct ast_node *node);
static void traverse_var_init_node(struct ast_node *node);
static void traverse_arg_init_node(struct ast_node *node);
static void traverse_array_init_node(struct ast_node *node);

static struct symbol_table *top;
/* Table of nodes handlers */
static struct {
	node_type_t node_type;
	travers_handler_t handler;
} traversers[] = {
	{ NODE_TYPE_ID,		traverse_id_node },
	{ NODE_TYPE_SUM_OP,	traverse_op_node },
	{ NODE_TYPE_MUL_OP,	traverse_op_node },
	{ NODE_TYPE_REL_OP,	traverse_op_node },
	{ NODE_TYPE_OR_OP,	traverse_op_node },
	{ NODE_TYPE_AND_OP,	traverse_op_node },
	{ NODE_TYPE_CONST,	traverse_const_node },
	{ NODE_TYPE_ACCESS,	traverse_access_node },
	{ NODE_TYPE_ASSIGN,	traverse_assign_node },
	{ NODE_TYPE_CALL,	traverse_call_node },
#ifdef _MEM_AREA
	{ NODE_TYPE_MEM_AREA,	traverse_mem_area_node },
#endif
	{ NODE_TYPE_IF,		traverse_if_node },
	{ NODE_TYPE_FOR,	traverse_for_node },
	{ NODE_TYPE_WHILE,	traverse_while_node },
	{ NODE_TYPE_RETURN,	traverse_return_node },
	{ NODE_TYPE_BREAK,	traverse_break_node },
	{ NODE_TYPE_CONTINUE, 	traverse_continue_node },
	{ NODE_TYPE_FUNCTION, 	traverse_function_node },
	{ NODE_TYPE_SCOPE,	traverse_scope_node },
	{ NODE_TYPE_SCOPE_END, 	traverse_scope_end_node },
	{ NODE_TYPE_VAR_INIT,	traverse_var_init_node },
	{ NODE_TYPE_ARG_INIT,	traverse_arg_init_node },
	{ NODE_TYPE_ARRAY_INIT,	traverse_array_init_node },
	{ NODE_TYPE_NOT,	traverse_not_node },
	{ NODE_TYPE_UNKNOWN,	NULL }
};

static struct {
	data_type_t type;
	char	*name;
} data_types[] = {
	{ DATA_TYPE_INTEGER,	"int" },
	{ DATA_TYPE_REAL,	"float" },
	{ DATA_TYPE_CHAR,	"char" },
	{ DATA_TYPE_STRING,	"string" },
	{ DATA_TYPE_VOID,	"void" },
	{ DATA_TYPE_UNKNOWN,	NULL }
};

static struct {
	opcode_t code;
	char *str;
} opers[] = {
	{ OPCODE_ADD,	"+" },
	{ OPCODE_SUB,	"-" },
	{ OPCODE_MUL,	"*" },
	{ OPCODE_DIV,	"/" },
	{ OPCODE_MOD,	"%" },
	{ OPCODE_LT,	"<" },
	{ OPCODE_LE,	"<=" },
	{ OPCODE_GT,	">" },
	{ OPCODE_GE,	">=" },
	{ OPCODE_EQ,	"==" },
	{ OPCODE_NE,	"!=" },
	{ OPCODE_OR,	"||" },
	{ OPCODE_AND,	"&&" },
	{ OPCODE_UNKNOWN, NULL }
};

/* A pointer to printing function */
static print_out_t print_out;

/* Default printing function. */
static void print_out_default(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	vprintf(fmt, ap);
	va_end(ap);
}

static const char *data_type_to_str(data_type_t type)
{
	unsigned int i;

	for (i = 0; data_types[i].type != DATA_TYPE_UNKNOWN; i++) {
		if (data_types[i].type == type)
			return data_types[i].name;
	}

	return "unknown";
}

static const char *oper_to_str(opcode_t code)
{
	unsigned int i;

	for (i = 0; opers[i].code != OPCODE_UNKNOWN; i++) {
		if (opers[i].code == code)
			return opers[i].str;
	}

	return "?";
}

/* Handle an identifier node, print variable name. */
static void traverse_id_node(struct ast_node *node)
{
	struct ast_node_id *id;

	return_if_fail(node != NULL);

	id = (struct ast_node_id *)node;

	print_out("%s", id->name);
}

/* Handle a constant node, print value according to the type. */
static void traverse_const_node(struct ast_node *node)
{
	struct ast_node_const *_const;

	return_if_fail(node != NULL);

	_const = (struct ast_node_const *)node;
	switch (_const->data_type) {
	case DATA_TYPE_CHAR:
		print_out("\'%c\'", _const->ch);
		break;
	case DATA_TYPE_INTEGER:
		print_out("%d", _const->integer);
		break;
	case DATA_TYPE_REAL:
		print_out("%f", _const->real);
		break;
	case DATA_TYPE_STRING:
		print_out("\"%s\"", _const->str);
		break;
	default:
		xerror(1, "unknown constant node type");
	}
}

/* Handle an access node, print the array name and go throw all indexes. */
static void traverse_access_node(struct ast_node *node)
{
	struct ast_node_access *access;
	unsigned int i;

	return_if_fail(node != NULL);

	access = (struct ast_node_access *)node;
	print_out("%s", access->name);
	for (i = 0; i < access->nidxs; i++) {
		print_out("[");
		travers_tree(access->idxs[i]);
		print_out("]");
	}
}

/* Handle an assign node, go throw left and right children. */
static void traverse_assign_node(struct ast_node *node)
{
	struct ast_node_assign *assign;
	unsigned int i;

	return_if_fail(node != NULL);

	assign = (struct ast_node_assign *)node;

	for (i = 0; i < assign->nleft; i++) {
		travers_tree(assign->left[i]);
		print_out("=");
	}
	travers_tree(assign->right);

	if (assign->right->type != NODE_TYPE_ASSIGN)
		print_out(";\n");
}

/* Handle a function call node, print the function name, go throw arguments. */
static void traverse_call_node(struct ast_node *node)
{
	struct ast_node_func_call *call;
	unsigned int i;

	return_if_fail(node != NULL);

	call = (struct ast_node_func_call *)node;

	print_out("%s", call->name);
	print_out("(");

	for (i = 0; i < call->nargs; i++) {
		travers_tree(call->args[i]);
		if (i < call->nargs - 1)
			print_out(",");
	}

	print_out(")");
	/* Is the simple call or a part of another expresstion? */
	if (node->parent != NULL && node->parent->type == NODE_TYPE_STMT)
		print_out(";\n");
}

#ifdef _MEM_AREA
static void traverse_mem_area_node(struct ast_node *node)
{
	struct ast_node_mem_area *mem_area;

	return_if_fail(node != NULL);

	mem_area = (struct ast_node_mem_area *)node;

	print_out("$%c%d", mem_area->area, mem_area->offset);
	if (mem_area->is_unsign)
		print_out("U");
	print_out("%c", mem_area->mem_data_type);

	if (mem_area->idx != NULL) {
		print_out("[");
		travers_tree(mem_area->idx);
		print_out("]");
	}

	if (mem_area->nbits != -1)
		print_out(".B%d", mem_area->nbits);
}
#endif

/* Handle an operation node, print arithmetics in the infix notation. */
static void traverse_op_node(struct ast_node *node)
{
	struct ast_node_op *op;
	boolean_t brackets;

	return_if_fail(node != NULL);

	op = (struct ast_node_op *)node;
	brackets = FALSE;

	if (node->parent != NULL) {
		struct ast_node *parent = node->parent;

		switch (parent->type) {
		case NODE_TYPE_MUL_OP:
		case NODE_TYPE_SUM_OP:
		case NODE_TYPE_REL_OP:
		case NODE_TYPE_AND_OP:
		case NODE_TYPE_OR_OP:
			if (parent->type < AST_NODE(op)->type)
				brackets = TRUE;
			break;
		default:
			break;
		}
	}

	if (brackets)
		print_out("(");

	travers_tree(op->left);
	print_out("%s", oper_to_str(op->opcode));
	travers_tree(op->right);

	if (brackets)
		print_out(")");
	/* In case of empty expression */
	if (node->parent != NULL && node->parent->type == NODE_TYPE_STMT)
		print_out(";\n");
}

/* Handle an if node, go throw condition expression, body statement, and
 * else statement if it exists. */
static void traverse_if_node(struct ast_node *node)
{
	struct ast_node_if *if_node;

	return_if_fail(node != NULL);

	if_node = (struct ast_node_if *)node;

	print_out("if");
	print_out("(");
	travers_tree(if_node->expr);
	print_out(")");
	travers_tree(if_node->stmt);

	if (if_node->_else != NULL) {
		print_out("else ");
		travers_tree(if_node->_else);
	}
}

/* Handle a for node, go throw all expressions and a body statement. */
static void traverse_for_node(struct ast_node *node)
{
	struct ast_node_for *for_node;

	return_if_fail(node != NULL);

	for_node = (struct ast_node_for *)node;

	print_out("for");
	print_out("(");
	if (for_node->optexpr1 != NULL)
		travers_tree(for_node->optexpr1);

	print_out(";");
	if (for_node->optexpr2 != NULL)
		travers_tree(for_node->optexpr2);

	print_out(";");
	if (for_node->optexpr3 != NULL)
		travers_tree(for_node->optexpr3);

	print_out(")");
	travers_tree(for_node->stmt);
}

/* Handle a while node, go throw an expression and a body statement. */
static void traverse_while_node(struct ast_node *node)
{
	struct ast_node_while *while_node;

	return_if_fail(node != NULL);

	while_node = (struct ast_node_while *)node;

	print_out("while");
	print_out("(");
	travers_tree(while_node->expr);
	print_out(")");
	travers_tree(while_node->stmt);
}

/* Handle a return node. */
static void traverse_return_node(struct ast_node *node)
{
	struct ast_node_return *ret;

	return_if_fail(node != NULL);

	ret = (struct ast_node_return *)node;

	print_out("return");

	if (ret->ret_val_node != NULL) {
		print_out(" ");
		travers_tree(ret->ret_val_node);
	}

	print_out(";\n");
}

/* Handle a break node. */
static void traverse_break_node(struct ast_node *node)
{
	return_if_fail(node != NULL);

	print_out("break;\n");
}

/* Handle a continue node */
static void traverse_continue_node(struct ast_node *node)
{
	return_if_fail(node != NULL);

	print_out("continue;\n");
}

/* Handle not node */
static void traverse_not_node(struct ast_node *node)
{
	struct ast_node_not *not_node;

	return_if_fail(node != NULL);

	not_node = (struct ast_node_not *)node;

	print_out("!(");
	travers_tree(not_node->expr);
	print_out(")");
}

static void print_decl_name(struct symbol *symbol)
{
	unsigned int i;

	print_out("%s", symbol->name);
	if (symbol->type == SYMBOL_TYPE_ARRAY) {
		for (i = 0; i < SYMBOL_ARRAY(symbol)->ndims; i++)
			print_out("[%d]", SYMBOL_ARRAY(symbol)->dims[i]);
	}
}

/* Handle a variable initialization node. */
static void traverse_var_init_node(struct ast_node *node)
{
	struct ast_node_var_init *var_init;
	struct symbol *s;

	return_if_fail(node != NULL);

	var_init = (struct ast_node_var_init *)node;
	s = AST_INIT(var_init)->symbol;

	print_decl_name(s);
	print_out("=");
	travers_tree(var_init->expr);
	print_out(";\n");
}

/* Handle a variable initialization node. */
static void traverse_arg_init_node(struct ast_node *node)
{
	struct ast_node_var_init *arg_init;

	return_if_fail(node != NULL);

	arg_init = (struct ast_node_var_init *)node;

	print_out("=");
	travers_tree(arg_init->expr);
}

/* Print array initialization nodes. */
static void print_array_init_nodes(struct ast_node *node, unsigned int *total)
{
	struct ast_node_array_init *array_init;
	struct array_idx idx;
	struct symbol_array *s;
	unsigned int i, j;

	return_if_fail(node != NULL);

	array_init = (struct ast_node_array_init *)node;
	s = (struct symbol_array *)AST_INIT(array_init)->symbol;

	/* Now we are in the most right index. */
	for (j = 0; j < s->dims[s->ndims-1]; j++) {
		for (i = 0; i < array_init->ninitvals; i++) {

			idx = array_init->initvals[i];

			if (idx.nidx == *total) {
				travers_tree(idx.expr);

				if (j < s->dims[s->ndims-1]-1)
					print_out("%c", ',');
			}
		}
		(*total)++;
	}
}

/* A recursive function for printing a multi-dimension array initialization node. */
static void cycle(struct ast_node *node, unsigned int iter, unsigned int *total)
{
	struct ast_node_array_init *array_init;
	struct symbol_array *s;
	unsigned int i;

	return_if_fail(node != NULL);

	array_init = (struct ast_node_array_init *)node;
	s = (struct symbol_array *)AST_INIT(array_init)->symbol;

	for (i = 0; i < s->dims[iter]; i++) {
		print_out("{");
		/* Go deeper before the last cell. */
		if (iter != s->ndims - 2)
			cycle(node, iter+1, total);
		else
			print_array_init_nodes(node, total);

		print_out("}");

		if (i < s->dims[iter] - 1)
			print_out(",");
	}
}

/* Handle an array initialization node. */
static void traverse_array_init_node(struct ast_node *node)
{
	struct ast_node_array_init *array_init;
	struct symbol_array *s;
	unsigned int total;

	return_if_fail(node != NULL);

	array_init = (struct ast_node_array_init *)node;
	s = (struct symbol_array *)AST_INIT(array_init)->symbol;

	print_decl_name(SYMBOL(s));
	print_out("={");

	total = 0;
	/* One dimension array. */
	if (s->ndims == 1)
		print_array_init_nodes(node, &total);
	else
		cycle(node, 0, &total);

	print_out("};\n");
}

static void print_storage_klass(struct symbol *s)
{
	switch(s->klass) {
	case SYMBOL_VISIB_GLOBAL:
		print_out("");
		break;
	case SYMBOL_VISIB_MODULE:
		print_out("module ");
		break;
	case SYMBOL_VISIB_EXTERN:
		print_out("extern ");
		break;
	default:
		/* Shut up gcc. */
		break;
	}
}

static void handle_variables(struct symbol **vars, unsigned int size)
{
	unsigned int i;

	for (i = 0; i < size; i++) {
		print_storage_klass(vars[i]);
		print_out("%s ", data_type_to_str(vars[i]->decl_type));
		print_decl_name(vars[i]);
		print_out(";\n");
	}
}

/* Handle variables and arrays declarations in this namespace scope. */
static void traverse_scope_node(struct ast_node *node)
{
	struct symbol **symbols;
	struct ast_node_scope *scope;
	unsigned int size;

	return_if_fail(node != NULL);

	scope = (struct ast_node_scope *)node;
	/* It is the first symbol table, it is the global table. */
	if (top != NULL)
		print_out("{\n");

	top = scope->table;

	symbols = symbol_table_extract_variables(scope->table, &size);
	if (symbols != NULL) {
		handle_variables(symbols, size);
		xfree(symbols);
	}

	travers_tree(scope->stmt);
}

/* Handle a scope end. */
static void traverse_scope_end_node(struct ast_node *node)
{
	struct symbol_table *global;

	return_if_fail(node != NULL);
	
	global = symbol_table_get_global_table();
	if (global != top)
		print_out("}\n");

	top = top->prev;
}

static void print_arg_attr(struct symbol_arg *arg)
{
	if (arg->is_fmt)
		print_out("fmt ");
	
	if (arg->is_ref)
		print_out("byref ");
}

/* Handle a function node, explore a function tree. */
static void traverse_function_node(struct ast_node *node)
{
	struct ast_node_func *func;
	struct symbol *s;
	unsigned int i;
	ret_t ret;

	return_if_fail(node != NULL);
	return_if_fail(top != NULL);

	func = (struct ast_node_func *)node;
	ret = hash_table_lookup(top->scope, func->name, (void **)&s);
	if (ret != ret_ok)
		xerror(1, "can't locate function symbol %s", func->name);

	print_out("\n");
	print_storage_klass(s);
	print_out("%s %s", data_type_to_str(s->decl_type), s->name);
	print_out("(");

	for (i = 0; i < SYMBOL_FUNC(s)->nargs; i++) {
		struct symbol_arg *arg = SYMBOL_FUNC(s)->args[i];
		print_arg_attr(arg);
		print_out("%s ", data_type_to_str(SYMBOL(arg)->decl_type));
		print_out("%s", SYMBOL(arg)->name);

		if (arg->initval != NULL) 
			travers_tree(arg->initval);

		if (i < SYMBOL_FUNC(s)->nargs-1)
			print_out(",");
	}

	print_out(")");
	travers_tree(func->func_tree);
}

/* The main traversal routine, some nodes are just containers and we explore
 * child and next nodes. */
static void travers_tree(struct ast_node *node)
{
	unsigned int i;

	return_if_fail(node != NULL);

	for (i = 0; traversers[i].node_type != NODE_TYPE_UNKNOWN; i++) {
		if (node->type == traversers[i].node_type) {
			traversers[i].handler(node);
			return;
		}
	}

	if (node->child != NULL)
		travers_tree(node->child);

	if (node->next != NULL)
		travers_tree(node->next);
}

/* The entry point for the conversion the tree back to the text. */
void tree2text(struct ast_node_root *root, print_out_t func)
{
	return_if_fail(root != NULL);

	if (func != NULL)
		print_out = func;
	else
		print_out = print_out_default;

	travers_tree(AST_NODE(root));
}

